@@ -9,8 +9,6 @@
  * License as published by the Free Software Foundation; either
  * version 2 of the License, or (at your option) any later version.
  *
-
-
  * This library is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
@@ -41,15 +39,19 @@
 #include <X11/cursorfont.h>
 #include <X11/extensions/XTest.h>
 #include <math.h>
+
+#undef ZOOM_REGION_DEBUG
+
 #include "zoom-region.h"
 #include "zoom-region-private.h"
 #include "magnifier.h" /* needed to access parent data */
 #include "magnifier-private.h" /* needed to access parent data */
 
-#undef DEBUG_CLIENT_CALLS
+#define DEBUG_CLIENT_CALLS
 
 #ifdef DEBUG_CLIENT_CALLS
-#define DBG(a) a
+static gboolean client_debug = FALSE;
+#define DBG(a) if (client_debug) { (a); }
 #else
 #define DBG(a) 
 #endif
@@ -213,6 +215,7 @@ static int  zoom_region_process_updates (gpointer data);
 static void zoom_region_paint (ZoomRegion *zoom_region, GdkRectangle *rect);
 static void zoom_region_paint_pixmap (ZoomRegion *zoom_region, GdkRectangle *rect);
 static int  zoom_region_update_pointer_timeout (gpointer data);
+static void zoom_region_recompute_exposed_viewport (ZoomRegion *zoom_region);
 
 void
 reset_timing_stats()
@@ -668,11 +671,21 @@ static GdkRectangle
 zoom_region_clip_to_window (ZoomRegion *zoom_region,
 			    GdkRectangle area)
 {
-	GdkRectangle window_rect = {0, 0, 0, 0};
+	GdkRectangle window_rect;
+
+	/* we can just return ATM because _rectangle_clip_to_rectangle is unimplemented now */
+
+	return area;
+
 	if (zoom_region->priv->w->window)
 		gdk_drawable_get_size (GDK_DRAWABLE (zoom_region->priv->w->window),
 				       &window_rect.x,
 				       &window_rect.y);
+	else 
+	{
+		window_rect.x = 0;
+		window_rect.y = 0;
+	}
 	return _rectangle_clip_to_rectangle (area, window_rect);
 }
 
@@ -739,22 +752,33 @@ zoom_region_queue_update (ZoomRegion *zoom_region,
 	GdkRectangle *rect =
 		g_new0 (GdkRectangle, 1);
 	*rect = update_rect;
+
+#ifdef ZOOM_REGION_DEBUG
+	g_assert (zoom_region->alive);
+#endif
 	DEBUG_RECT ("queueing update", *rect);
+
 	zoom_region->priv->q =
-	g_list_prepend (zoom_region->priv->q, rect);
-	g_idle_add_full (G_PRIORITY_DEFAULT_IDLE,
-			 zoom_region_process_updates,
-			 zoom_region,
-			 NULL);
+		g_list_prepend (zoom_region->priv->q, rect);
+	if (zoom_region->priv && zoom_region->priv->update_handler_id == 0)
+		zoom_region->priv->update_handler_id = 
+			g_idle_add_full (G_PRIORITY_DEFAULT_IDLE,
+					 zoom_region_process_updates,
+					 zoom_region,
+					 NULL);
 }
 
 static void
 zoom_region_update_current (ZoomRegion *zoom_region)
 {
-	zoom_region_update (zoom_region,
-			    zoom_region_source_rect_from_view_bounds (
-				    zoom_region,
-				    &zoom_region->priv->exposed_viewport));
+#ifdef ZOOM_REGION_DEBUG
+	g_assert (zoom_region->alive);
+#endif
+	if (zoom_region->priv && GDK_IS_DRAWABLE (zoom_region->priv->pixmap))
+	    zoom_region_update (zoom_region,
+				zoom_region_source_rect_from_view_bounds (
+					zoom_region,
+					&zoom_region->priv->exposed_viewport));
 }
 
 static GdkRectangle
@@ -785,7 +809,10 @@ zoom_region_unpaint_crosswire_cursor (ZoomRegion *zoom_region,
 	GdkRectangle vline_rect, hline_rect;
 	GdkPoint cursor_pos;
 
-	if (magnifier->crosswire_size <= 0) return;
+#ifdef ZOOM_REGION_DEBUG
+	g_assert (zoom_region->alive);
+#endif
+	if (!magnifier || magnifier->crosswire_size <= 0) return;
 
 	cursor_pos = zoom_region->priv->last_drawn_crosswire_pos;
 	vline_rect.x = cursor_pos.x - magnifier->crosswire_size/2;
@@ -816,6 +843,9 @@ zoom_region_paint_crosswire_cursor (ZoomRegion *zoom_region, GdkRectangle *clip_
 	int x_left_clip = 0, x_right_clip = 0, y_top_clip = 0, y_bottom_clip = 0;
 	int csize;
 	
+#ifdef ZOOM_REGION_DEBUG
+	g_assert (zoom_region->alive);
+#endif
 	if (!(magnifier &&
 	      zoom_region->priv->w->window &&
 	      GDK_IS_DRAWABLE (zoom_region->priv->w->window) &&
@@ -867,6 +897,8 @@ zoom_region_paint_crosswire_cursor (ZoomRegion *zoom_region, GdkRectangle *clip_
 
 	rect.x = zoom_region->priv->last_cursor_pos.x;
 	rect.y = zoom_region->priv->last_cursor_pos.y;
+	rect.width = 0;
+	rect.height = 0;
 	rect = zoom_region_view_rect_from_source_rect (zoom_region, rect);
 	if (clip_rect) gdk_gc_set_clip_rectangle (zoom_region->priv->crosswire_gc, clip_rect);
 	else gdk_gc_set_clip_rectangle (zoom_region->priv->crosswire_gc, NULL);
@@ -923,6 +955,9 @@ zoom_region_paint_crosswire_cursor (ZoomRegion *zoom_region, GdkRectangle *clip_
 static void
 zoom_region_unpaint_cursor (ZoomRegion *zoom_region, GdkRectangle *clip_rect)
 {
+#ifdef ZOOM_REGION_DEBUG
+	g_assert (zoom_region->alive);
+#endif
 	zoom_region_paint_pixmap (zoom_region, &zoom_region->priv->cursor_backing_rect);
 }
 
@@ -935,6 +970,9 @@ zoom_region_paint_cursor (ZoomRegion *zoom_region,
 	GdkRectangle fullscreen;
 	Magnifier *magnifier = zoom_region->priv->parent;
 	rect = zoom_region_cursor_rect (zoom_region);
+#ifdef ZOOM_REGION_DEBUG
+	g_assert (zoom_region->alive);
+#endif
 	if (clip_rect == NULL)
 	{
 		fullscreen = zoom_region_rect_from_bounds (zoom_region,
@@ -1023,14 +1061,19 @@ zoom_region_coalesce_updates (ZoomRegion *zoom_region)
 	GList *q;
 	int lookahead_n = 4; /* 'distance' to look ahead in queue */
 	int max_qlen = 50;
-	if (g_list_length (zoom_region->priv->q) > max_qlen)
+
+	if (zoom_region->priv && zoom_region->priv->q && g_list_length (zoom_region->priv->q) > max_qlen)
 	{
+		g_list_free (zoom_region->priv->q);
 		zoom_region->priv->q = NULL; /* just discard and update everything */
 		/* CAUTION: this can be an expensive operation! */
 		zoom_region_queue_update (zoom_region, zoom_region_rect_from_bounds
 			(zoom_region, &zoom_region->priv->source_area));
 	}
-	else if ((g_list_length (zoom_region->priv->q) > 1) && can_coalesce)
+	else 
+
+        if (zoom_region->priv && zoom_region->priv->q && 
+	    (g_list_length (zoom_region->priv->q) > 1) && can_coalesce)
 	{		
 		q = g_list_reverse (g_list_copy (zoom_region->priv->q));
 		if (q)
@@ -1055,6 +1098,10 @@ zoom_region_paint_border (ZoomRegion *zoom_region,
 			  GdkRectangle *area)
 {
 	GdkColor color;
+
+#ifdef ZOOM_REGION_DEBUG
+	g_assert (zoom_region->alive);
+#endif
 	if ((zoom_region->border_size > 0) && (zoom_region->priv->w->window))
 	{
 		if (!zoom_region->priv->border_gc)
@@ -1085,9 +1132,15 @@ static void
 zoom_region_paint_pixmap (ZoomRegion *zoom_region,
 			  GdkRectangle *area)
 {
-        if (!zoom_region->priv->w->window) return;
+#ifdef ZOOM_REGION_DEBUG
+	g_assert (zoom_region->alive);
+#endif
+	g_assert (zoom_region->priv);
+	g_assert (zoom_region->priv->w);
+
+        if (!GDK_IS_DRAWABLE (zoom_region->priv->w->window)) return;
 	if (zoom_region->priv->default_gc == NULL) 
-		zoom_region->priv->default_gc = gdk_gc_new(zoom_region->priv->w->window);
+		zoom_region->priv->default_gc = gdk_gc_new (zoom_region->priv->w->window);
 
 	if (zoom_region->priv->pixmap && GDK_IS_DRAWABLE (zoom_region->priv->w->window))
 	{
@@ -1111,6 +1164,10 @@ zoom_region_paint (ZoomRegion *zoom_region,
 		   GdkRectangle *area)
 {
 	GdkRectangle paint_area;
+
+#ifdef ZOOM_REGION_DEBUG
+	g_assert (zoom_region->alive);
+#endif
 	DEBUG_RECT ("painting (clipped)", *area);
 	paint_area = zoom_region_clip_to_window (zoom_region, *area);
 	zoom_region_paint_border (zoom_region, area);
@@ -1122,6 +1179,9 @@ zoom_region_paint (ZoomRegion *zoom_region,
 static ZoomRegionPixmapCreationError
 zoom_region_create_pixmap (ZoomRegion *zoom_region)
 {
+#ifdef ZOOM_REGION_DEBUG
+	g_assert (zoom_region->alive);
+#endif
 	if (zoom_region->priv->w && GDK_IS_DRAWABLE (zoom_region->priv->w->window))
 	{
 		long width = (zoom_region->priv->source_area.x2 -
@@ -1141,10 +1201,28 @@ zoom_region_create_pixmap (ZoomRegion *zoom_region)
 			return ZOOM_REGION_ERROR_TOO_BIG;
 		}
 
-		zoom_region_update (zoom_region,
-				    zoom_region_source_rect_from_view_bounds (
+		zoom_region_recompute_exposed_viewport (zoom_region);
+#ifdef ZOOM_REGION_DEBUG
+		g_message ("create-pixmap-update: %d,%d - %d,%d",
+			   zoom_region->priv->exposed_viewport.x1,
+			   zoom_region->priv->exposed_viewport.y1,
+			   zoom_region->priv->exposed_viewport.x2,
+			   zoom_region->priv->exposed_viewport.y2);
+#endif
+
+			    DEBUG_RECT("viewport", zoom_region_source_rect_from_view_bounds
+					    (zoom_region, &zoom_region->priv->exposed_viewport));
+			    DEBUG_RECT("source", zoom_region_rect_from_bounds
+					    (zoom_region, &((Magnifier*)zoom_region->priv->parent)->source_bounds));
+
+			    zoom_region_update (zoom_region,
+/*				    zoom_region_source_rect_from_view_bounds (
 					    zoom_region,
 					    &zoom_region->priv->exposed_viewport));
+*/
+				    zoom_region_rect_from_bounds 
+				    (zoom_region, 
+				     &((Magnifier *)zoom_region->priv->parent)->source_bounds));
 		return ZOOM_REGION_ERROR_NONE;
 	}
 
@@ -1159,16 +1237,22 @@ zoom_region_expose_handler (GtkWindow * w,
 	ZoomRegion *zoom_region = data;
 	DEBUG_RECT ("expose", event->area);
 
+#ifdef ZOOM_REGION_DEBUG
+	g_assert (zoom_region->alive);
+#endif
 	if (zoom_region->priv->pixmap == NULL)
 	{
+	        ZoomRegionPixmapCreationError ret; 
 		/* TODO: scale down if this fails here */
-		while (zoom_region_create_pixmap (zoom_region) ==
+		while ((ret = zoom_region_create_pixmap (zoom_region)) ==
 		    ZOOM_REGION_ERROR_TOO_BIG) {
 			zoom_region->xscale -= 1.0;
 			zoom_region->yscale -= 1.0;
 			zoom_region->priv->pixmap = NULL;
 			g_warning ("Scale factor too big to fit in memory; shrinking.");
 		}
+		if (ret == ZOOM_REGION_ERROR_NO_TARGET_DRAWABLE) 
+		    g_warning ("create-pixmap: no target drawable");
 	}
 	zoom_region_paint (zoom_region, &event->area);
 }
@@ -1176,6 +1260,9 @@ zoom_region_expose_handler (GtkWindow * w,
 static void zoom_region_update_cursor (ZoomRegion *zoom_region, int dx, int dy,
 				       GdkRectangle *clip_rect)
 {
+#ifdef ZOOM_REGION_DEBUG
+	g_assert (zoom_region->alive);
+#endif
 	zoom_region_unpaint_crosswire_cursor (zoom_region, clip_rect);
 	zoom_region_unpaint_cursor (zoom_region, clip_rect);
 	zoom_region->priv->cursor_backing_rect.x += dx;
@@ -1184,7 +1271,8 @@ static void zoom_region_update_cursor (ZoomRegion *zoom_region, int dx, int dy,
 	zoom_region->priv->last_drawn_crosswire_pos.y += dy;
 	zoom_region_paint_cursor (zoom_region, clip_rect);
 	zoom_region_paint_crosswire_cursor (zoom_region, clip_rect);
-	gdk_display_sync (gdk_drawable_get_display (zoom_region->priv->w->window));
+	if (GTK_IS_WIDGET (zoom_region->priv->w) && GDK_IS_WINDOW (zoom_region->priv->w->window))
+	    gdk_display_sync (gdk_drawable_get_display (zoom_region->priv->w->window));
 }
 
 static gboolean
@@ -1194,17 +1282,22 @@ zoom_region_calculate_scroll_rects (ZoomRegion *zoom_region,
 				    GdkRectangle *expose_rect_h,
 				    GdkRectangle *expose_rect_v)
 {
-	GdkWindow *window;
-	GdkRectangle rect;
+	GdkWindow *window = NULL;
+	GdkRectangle rect = {0, 0, 0, 0};
+	gboolean retval = TRUE;
+
+#ifdef ZOOM_REGION_DEBUG
+	g_assert (zoom_region->alive);
+#endif
 	rect.x = 0;
 	rect.y = 0;
 	if (zoom_region && zoom_region->priv->w &&
 	    zoom_region->priv->w->window)
 		window = zoom_region->priv->w->window;
 	else
-		return FALSE;
+		retval = FALSE;
 	if (!window)
-		return FALSE;
+		retval = FALSE;
 
 	if (window != NULL)
 	  gdk_drawable_get_size (GDK_DRAWABLE (window),
@@ -1213,8 +1306,8 @@ zoom_region_calculate_scroll_rects (ZoomRegion *zoom_region,
 
 	if ((ABS (dx) >= rect.width) || (ABS (dy) >= rect.height)) {
 		*scroll_rect = rect;
-		fprintf (stderr, "deltas too big to scroll\n");
-		return FALSE;
+		DBG(fprintf (stderr, "deltas too big to scroll\n"));
+		retval = FALSE;
 	}
 	
 	scroll_rect->x = MAX (0, dx);
@@ -1232,7 +1325,7 @@ zoom_region_calculate_scroll_rects (ZoomRegion *zoom_region,
 	expose_rect_v->width = rect.width - scroll_rect->width;
 	expose_rect_v->height = scroll_rect->height;
 
-	return TRUE;
+	return retval;
 }
 
 static void
@@ -1242,6 +1335,10 @@ zoom_region_scroll_fast (ZoomRegion *zoom_region, int dx, int dy,
 			 GdkRectangle *expose_rect_v)
 {
 	GdkWindow *window;
+
+#ifdef ZOOM_REGION_DEBUG
+	g_assert (zoom_region->alive);
+#endif
 	if (zoom_region->priv->w && zoom_region->priv->w->window)
 		window = zoom_region->priv->w->window;
 	else {
@@ -1268,6 +1365,9 @@ zoom_region_scroll_smooth (ZoomRegion *zoom_region, int dx, int dy,
 	GdkWindow *window = NULL;
 	GdkRectangle window_rect;
 
+#ifdef ZOOM_REGION_DEBUG
+	g_assert (zoom_region->alive);
+#endif
 	if (zoom_region->priv->w && GDK_IS_DRAWABLE (zoom_region->priv->w->window))
 		window = zoom_region->priv->w->window;
 	else
@@ -1290,6 +1390,9 @@ zoom_region_scroll (ZoomRegion *zoom_region, int dx, int dy)
 	GdkRectangle scroll_rect, expose_rect_h, expose_rect_v;
 	gboolean can_scroll;
 
+#ifdef ZOOM_REGION_DEBUG
+	g_assert (zoom_region->alive);
+#endif
 	if (timing_test) {
 		mag_timing.num_line_samples++;
 		mag_timing.dx = abs(dx);
@@ -1366,7 +1469,6 @@ zoom_region_recompute_exposed_bounds (ZoomRegion *zoom_region)
 static void
 zoom_region_set_cursor_pos (ZoomRegion *zoom_region, int x, int y)
 {
-
 	if (zoom_region->priv)
 	{
 		zoom_region->priv->last_cursor_pos.x = x;
@@ -1395,10 +1497,14 @@ zoom_region_update_pointer (ZoomRegion *zoom_region, gboolean draw_cursor)
 	gint mouse_x_return, mouse_y_return;
 	guint mask_return;
 
+#ifdef ZOOM_REGION_DEBUG
+	g_assert (zoom_region->alive);
+#endif
 	if (!zoom_region->priv || !zoom_region->priv->parent)
-	  return FALSE; 
+	      return FALSE; 
 
 	magnifier = zoom_region->priv->parent;
+
 	/* TODO: there's really no reason we should be using magnifier->priv->root here */
 	if (magnifier && magnifier->priv && magnifier_get_root (magnifier))
 	{
@@ -1411,17 +1517,14 @@ zoom_region_update_pointer (ZoomRegion *zoom_region, gboolean draw_cursor)
 		if (zoom_region->priv->last_cursor_pos.x != mouse_x_return
 		    || zoom_region->priv->last_cursor_pos.y != mouse_y_return)
 		{
-			/*
-			 * dx and dy offsets are with respect to stored cursor coords;
-			 * these only should be nonzero if the window coord system has
-			 * changed, for instance via a gdk_window_scroll.
-			 */
 			zoom_region_set_cursor_pos (zoom_region,
 						    mouse_x_return, mouse_y_return);
 			if (draw_cursor)
 			{
 				GdkRectangle paint_area, *clip = NULL;
-				if (GDK_IS_DRAWABLE (zoom_region->priv->w->window))
+
+				if (GTK_IS_WIDGET (zoom_region->priv->w) && 
+				    GDK_IS_DRAWABLE (zoom_region->priv->w->window))
 				{
 					gdk_drawable_get_size (
 						GDK_DRAWABLE (
@@ -1445,14 +1548,17 @@ static int
 zoom_region_update_pointer_idle (gpointer data)
 {
 	ZoomRegion *zoom_region = (ZoomRegion *) data;
+
 	if (zoom_region_update_pointer (zoom_region, TRUE))
 	        return TRUE;
 	else {
-	        g_timeout_add_full (G_PRIORITY_DEFAULT,
-				    100,
-				    zoom_region_update_pointer_timeout,
-				    zoom_region,
-				    NULL);
+		if (zoom_region->priv)
+			zoom_region->priv->update_pointer_id =
+			    g_timeout_add_full (G_PRIORITY_DEFAULT,
+						100,
+						zoom_region_update_pointer_timeout,
+						zoom_region,
+						NULL);
                 return FALSE;
 	}
 }
@@ -1462,13 +1568,14 @@ zoom_region_update_pointer_timeout (gpointer data)
 {
 	ZoomRegion *zoom_region = data;
 
-	if (zoom_region_update_pointer (zoom_region, TRUE)) {
+	if (zoom_region->priv && zoom_region_update_pointer (zoom_region, TRUE)) {
+	    zoom_region->priv->update_pointer_id =
 	        g_idle_add_full (G_PRIORITY_HIGH_IDLE,
 				 zoom_region_update_pointer_idle,
 				 data,
 				 NULL);
 		return FALSE;
-	} else
+	} else 
 		return TRUE;
 }
 
@@ -1478,6 +1585,9 @@ zoom_region_moveto (ZoomRegion *zoom_region,
 {
 	long dx = x * zoom_region->xscale - zoom_region->priv->exposed_bounds.x1;
 	long dy = y * zoom_region->yscale - zoom_region->priv->exposed_bounds.y1;
+#ifdef ZOOM_REGION_DEBUG
+	g_assert (zoom_region->alive);
+#endif
 /* fprintf (stderr, "moveto %ld %ld\n", x, y); */
 
 	mag_timing.dx = 0;
@@ -1545,6 +1655,10 @@ zoom_region_get_source_subwindow (ZoomRegion *zoom_region,
 	int i, j, width, height;
 	Magnifier *magnifier = zoom_region->priv->parent;
 	GdkPixbuf *subimage = NULL;
+
+#ifdef ZOOM_REGION_DEBUG
+	g_assert (zoom_region->alive);
+#endif
 	width = gdk_screen_get_width (
 		gdk_display_get_screen (magnifier->source_display,
 					magnifier->source_screen_num));
@@ -1553,11 +1667,13 @@ zoom_region_get_source_subwindow (ZoomRegion *zoom_region,
 					magnifier->source_screen_num));
 
 	if ((bounds.width <= 0) || (bounds.height <= 0))
+	{
 		return NULL;
+	}
 	
-	/* TESTING ONLY */
 	if (!zoom_region->priv->source_drawable)
 	{
+		/* TESTING ONLY */
 		if (zoom_region->priv->test) {
 			GdkImage *test_image = NULL;
 
@@ -1609,6 +1725,8 @@ zoom_region_get_source_subwindow (ZoomRegion *zoom_region,
 						 0,
 						 bounds.width,
 						 bounds.height);
+
+	/* TODO: blank the region overlapped by the target display if source == target */
 	
 	if (!subimage)
 		_debug_announce_rect ("update of invalid subregion!\n", bounds);
@@ -1646,11 +1764,16 @@ zoom_region_update (ZoomRegion *zoom_region,
 {
 	GdkPixbuf *subimage;
 	GdkRectangle source_rect;
+
+#ifdef ZOOM_REGION_DEBUG
+	g_assert (zoom_region->alive);
+#endif
 	DEBUG_RECT ("unclipped update rect", update_rect);
 	source_rect = zoom_region_clip_to_source (zoom_region, update_rect);
 	DEBUG_RECT ("clipped to source", source_rect);
 	source_rect = zoom_region_clip_to_exposed_target (zoom_region, source_rect);
 	DEBUG_RECT ("update rect clipped to exposed target", source_rect); 
+
 	subimage = zoom_region_get_source_subwindow (zoom_region, source_rect);
 
 	if (subimage && zoom_region->priv->w && zoom_region->priv->w->window)
@@ -1695,7 +1818,7 @@ zoom_region_update (ZoomRegion *zoom_region,
 				     0,
 				     0);
 		else
-		    g_warning ("updating non-drawable pixmap");
+		    g_warning ("updating non-drawable pixmap: region %p", zoom_region);
 #else
 		gdk_pixbuf_render_to_drawable (zoom_region->priv->scaled_pixbuf,
 					       zoom_region->priv->pixmap,
@@ -1758,19 +1881,22 @@ zoom_region_init_window (ZoomRegion *zoom_region)
 {
 	GtkFixed *parent;
 	GtkWidget *zoomer, *border;
-	fprintf (stderr, "window not yet created...\n");
+	DBG(fprintf (stderr, "window not yet created...\n"));
 	parent = GTK_FIXED (
 		((Magnifier *)zoom_region->priv->parent)->priv->canvas);
 	zoomer = gtk_drawing_area_new ();
 	border = gtk_fixed_new ();
 	zoom_region->priv->border = border;
 	zoom_region->priv->w = zoomer;
+
+#ifdef ZOOM_REGION_DEBUG
+	g_assert (zoom_region->alive);
+#endif
 	gtk_widget_set_size_request (GTK_WIDGET (zoomer),
 				     zoom_region->viewport.x2 -
 				     zoom_region->viewport.x1 - zoom_region->border_size * 2,
 				     zoom_region->viewport.y2 -
 				     zoom_region->viewport.y1 - zoom_region->border_size * 2);
-	fprintf (stderr, "putting into fixed\n");
 	gtk_fixed_put (parent, border,
 		       zoom_region->viewport.x1,
 		       zoom_region->viewport.y1);
@@ -1780,13 +1906,12 @@ zoom_region_init_window (ZoomRegion *zoom_region)
 	gtk_widget_show (GTK_WIDGET (border));
 	gtk_widget_show (GTK_WIDGET (zoomer));
 	gtk_widget_show (GTK_WIDGET (parent));
-	fprintf (stderr, "shown\n");
 	zoom_region->priv->expose_handler_id =
 		g_signal_connect (G_OBJECT (zoom_region->priv->w),
 			    "expose_event",
 			    G_CALLBACK (zoom_region_expose_handler),
 			    zoom_region);
-	fprintf (stderr, "New window created\n");
+	DBG(fprintf (stderr, "New window created\n"));
 	gtk_widget_show (GTK_WIDGET (zoom_region->priv->w));
 }
 
@@ -1794,11 +1919,15 @@ static int
 zoom_region_process_updates (gpointer data)
 {
 	ZoomRegion *zoom_region = (ZoomRegion *) data;
+
 	/* TODO: lock the queue when copying it? */
 	zoom_region_coalesce_updates (zoom_region);
+
 	if (zoom_region->priv->q != NULL) {
 		GList *last = g_list_last (zoom_region->priv->q);
-/*		fprintf (stderr, "qlen=%d\n", g_list_length (zoom_region->priv->q));*/
+#ifdef ZOOM_REGION_DEBUG
+		fprintf (stderr, "qlen=%d\n", g_list_length (zoom_region->priv->q));
+#endif
 		if (last) {
 			zoom_region->priv->q = g_list_remove_link (zoom_region->priv->q,
 								   last);
@@ -1812,7 +1941,11 @@ zoom_region_process_updates (gpointer data)
 		return TRUE;
 	}
 	else 
+	{
+		if (zoom_region->priv) 
+			zoom_region->priv->update_handler_id = 0;
 		return FALSE;
+	}
 }
 
 void
@@ -2120,14 +2253,15 @@ static void
 zoom_region_set_viewport (ZoomRegion *zoom_region,
 			  const GNOME_Magnifier_RectBounds *viewport)
 {
+#ifdef ZOOM_REGION_DEBUG
+	g_assert (zoom_region->alive);
+#endif
 	zoom_region->viewport = *viewport;
-#define DEBUG
 #ifdef DEBUG
 	fprintf (stderr, "Setting viewport %d,%d - %d,%d\n",
 		 (int) viewport->x1, (int) viewport->y1,
 		 (int) viewport->x2, (int) viewport->y2);
 #endif
-#undef DEBUG
 	zoom_region_recompute_exposed_viewport (zoom_region);
 	zoom_region_align (zoom_region);
 	if (!zoom_region->priv->w) {
@@ -2159,6 +2293,9 @@ zoom_region_get_property (BonoboPropertyBag *bag,
 {
 	ZoomRegion *zoom_region = user_data;
 
+#ifdef ZOOM_REGION_DEBUG
+	g_assert (zoom_region->alive);
+#endif
 	DBG (fprintf (stderr, "Get zoom-region property: %s\n", prop_names[arg_id]));
 
 	switch (arg_id) {
@@ -2233,6 +2370,9 @@ zoom_region_set_property (BonoboPropertyBag *bag,
 	ZoomRegion *zoom_region = user_data;
 	GNOME_Magnifier_RectBounds bounds;
 
+#ifdef ZOOM_REGION_DEBUG
+	g_assert (zoom_region->alive);
+#endif
 	DBG (fprintf (stderr, "Set zoom-region property: %s\n", prop_names[arg_id]));
 
 	switch (arg_id) {
@@ -2324,6 +2464,10 @@ static int
 zoom_region_process_pending (gpointer data)
 {
 	ZoomRegion *zoom_region = (ZoomRegion *) data;
+
+#ifdef ZOOM_REGION_DEBUG
+	g_assert (zoom_region->alive);
+#endif
 	zoom_region_align (zoom_region);
 	return FALSE;
 }
@@ -2338,7 +2482,7 @@ zoom_region_pan_test (gpointer data)
 	CORBA_Environment ev;
 	static int counter = 0;
 	static gboolean finished_update = !TRUE;
-    static float last_pixels_at_speed = -1;
+        static float last_pixels_at_speed = -1;
 	float pixels_at_speed;
 	float total_time;
 	int screen_height, height;
@@ -2426,6 +2570,9 @@ impl_zoom_region_set_roi (PortableServer_Servant servant,
 	ZoomRegion *zoom_region =
 		ZOOM_REGION (bonobo_object_from_servant (servant));
 
+#ifdef ZOOM_REGION_DEBUG
+	g_assert (zoom_region->alive);
+#endif
 	DBG (fprintf (stderr, "Set ROI: \t%d,%d %d,%d\n", 
 		      bounds->x1, bounds->y1, bounds->x2, bounds->y2));
 
@@ -2470,6 +2617,7 @@ impl_zoom_region_set_roi (PortableServer_Servant servant,
 		}
 
 		/* Set idle handler to process this pending update when possible */
+
 		pending_idle_handler = g_idle_add_full (GDK_PRIORITY_REDRAW + 2,
 			zoom_region_process_pending, zoom_region, NULL);
 
@@ -2492,6 +2640,10 @@ impl_zoom_region_set_mag_factor (PortableServer_Servant servant,
 {
 	ZoomRegion *zoom_region =
 		ZOOM_REGION (bonobo_object_from_servant (servant));
+
+#ifdef ZOOM_REGION_DEBUG
+	g_assert (zoom_region->alive);
+#endif
 	CORBA_any *any;
 	double xs_old = zoom_region->xscale;
 	double ys_old = zoom_region->yscale;
@@ -2556,6 +2708,10 @@ impl_zoom_region_get_mag_factor (PortableServer_Servant servant,
 {
 	ZoomRegion *zoom_region =
 		ZOOM_REGION (bonobo_object_from_servant (servant));
+
+#ifdef ZOOM_REGION_DEBUG
+	g_assert (zoom_region->alive);
+#endif
 	*mag_factor_x = zoom_region->xscale;
 	*mag_factor_y = zoom_region->yscale;
 }
@@ -2566,6 +2722,10 @@ impl_zoom_region_get_properties (PortableServer_Servant servant,
 {
 	ZoomRegion *zoom_region =
 		ZOOM_REGION (bonobo_object_from_servant (servant));
+
+#ifdef ZOOM_REGION_DEBUG
+	g_assert (zoom_region->alive);
+#endif
 	return bonobo_object_dup_ref (
 		BONOBO_OBJREF (zoom_region->properties), ev);
 }
@@ -2577,9 +2737,15 @@ impl_zoom_region_mark_dirty (PortableServer_Servant servant,
 {
 	ZoomRegion *zoom_region =
 		ZOOM_REGION (bonobo_object_from_servant (servant));
+
+#ifdef ZOOM_REGION_DEBUG
+	g_assert (zoom_region->alive);
+#endif
 	DEBUG_RECT ("mark dirty", zoom_region_rect_from_bounds (
 			    zoom_region, roi_dirty) );
+
 	zoom_region_update_pointer (zoom_region, TRUE);
+	/* XXX ? should we clip here, or wait till process_updates? */
 	zoom_region_queue_update (zoom_region, 
 	  zoom_region_clip_to_target (zoom_region, 
 	      zoom_region_rect_from_bounds (zoom_region, roi_dirty)));
@@ -2591,6 +2757,10 @@ impl_zoom_region_get_roi (PortableServer_Servant servant,
 {
 	ZoomRegion *zoom_region =
 		ZOOM_REGION (bonobo_object_from_servant (servant));
+
+#ifdef ZOOM_REGION_DEBUG
+	g_assert (zoom_region->alive);
+#endif
 	return zoom_region->roi;
 }
 
@@ -2601,19 +2771,34 @@ impl_zoom_region_move_resize (PortableServer_Servant            servant,
 {
 	ZoomRegion *zoom_region =
 		ZOOM_REGION (bonobo_object_from_servant (servant));
+
+#ifdef ZOOM_REGION_DEBUG
+	g_assert (zoom_region->alive);
+#endif
 	zoom_region_set_viewport (zoom_region, viewport_bounds);
 }
 
+/* could be called multiple times... */
 static void
 zoom_region_do_dispose (ZoomRegion *zoom_region)
 {
-	if (zoom_region->priv->expose_handler_id) {
+	DBG(g_message ("disposing region %p", zoom_region));
+	if (zoom_region->priv && zoom_region->priv->expose_handler_id && 
+	    GTK_IS_WIDGET (zoom_region->priv->w)) {
 		g_signal_handler_disconnect (
 			zoom_region->priv->w,
 			zoom_region->priv->expose_handler_id);
 		zoom_region->priv->expose_handler_id = 0;
 	}
+	if (zoom_region->priv && zoom_region->priv->update_pointer_id)
+	    g_source_remove (zoom_region->priv->update_pointer_id);
+	if (zoom_region->priv && zoom_region->priv->update_handler_id)
+	    g_source_remove (zoom_region->priv->update_handler_id);
 	g_idle_remove_by_data (zoom_region);
+	
+#ifdef ZOOM_REGION_DEBUG
+	zoom_region->alive = FALSE;
+#endif
 }
 
 static void
@@ -2625,6 +2810,8 @@ impl_zoom_region_dispose (PortableServer_Servant servant,
 	zoom_region_do_dispose (zoom_region);
 }
 
+
+/* could be called multiple times */
 static void
 zoom_region_dispose (GObject *object)
 {
@@ -2640,6 +2827,7 @@ zoom_region_class_init (ZoomRegionClass *klass)
 {
 	GObjectClass * object_class = (GObjectClass *) klass;
 	POA_GNOME_Magnifier_ZoomRegion__epv *epv = &klass->epv;
+	parent_class = g_type_class_peek (BONOBO_TYPE_OBJECT); /* needed by BONOBO_CALL_PARENT! */
 
 	object_class->dispose = zoom_region_dispose;
 	object_class->finalize = zoom_region_finalize;
@@ -2654,6 +2842,9 @@ zoom_region_class_init (ZoomRegionClass *klass)
 	epv->dispose = impl_zoom_region_dispose;
 
 	reset_timing_stats();
+#ifdef DEBUG_CLIENT_CALLS
+	client_debug = (g_getenv ("MAG_CLIENT_DEBUG") != NULL);
+#endif
 }
 
 static void
@@ -2892,6 +3083,7 @@ static void
 zoom_region_private_init (ZoomRegionPrivate *priv)
 {
 	GdkRectangle rect = {0, 0, 0, 0};
+	GNOME_Magnifier_RectBounds rectbounds = {0, 0, 0, 0};
 	priv->parent = NULL;
 	priv->w = NULL;
 	priv->default_gc = NULL;
@@ -2902,7 +3094,6 @@ zoom_region_private_init (ZoomRegionPrivate *priv)
 	priv->source_pixbuf_cache = NULL;
 	priv->source_drawable = NULL;
 	priv->pixmap = NULL;
-	priv->parent = NULL;
 	priv->cursor_backing_rect = rect;
 	priv->cursor_backing_pixels = NULL;
 	priv->border_gc = NULL;
@@ -2913,11 +3104,18 @@ zoom_region_private_init (ZoomRegionPrivate *priv)
 	priv->last_cursor_pos.y = 0;
 	priv->last_drawn_crosswire_pos.x = 0;
 	priv->last_drawn_crosswire_pos.y = 0;
+	priv->exposed_bounds = rectbounds;
+	priv->exposed_viewport = rectbounds;
+	priv->source_area = rectbounds;
+	priv->update_pointer_id = 0;
+	priv->update_handler_id = 0;
 }
 
 static void
 zoom_region_init (ZoomRegion *zoom_region)
 {
+	DBG(g_message ("initializing region %p", zoom_region));
+
 	zoom_region_properties_init (zoom_region);
 	zoom_region->smooth_scroll_policy =
 		GNOME_Magnifier_ZoomRegion_SCROLL_SMOOTH;
@@ -2932,16 +3130,20 @@ zoom_region_init (ZoomRegion *zoom_region)
 	zoom_region->x_align_policy = GNOME_Magnifier_ZoomRegion_ALIGN_CENTER;
 	zoom_region->y_align_policy = GNOME_Magnifier_ZoomRegion_ALIGN_CENTER;
 	zoom_region->coalesce_func = _coalesce_update_rects;
-	bonobo_object_add_interface (BONOBO_OBJECT (zoom_region),
-				     BONOBO_OBJECT (zoom_region->properties));
 	zoom_region->priv = g_malloc (sizeof (ZoomRegionPrivate));
 	zoom_region_private_init (zoom_region->priv);
+	bonobo_object_add_interface (BONOBO_OBJECT (zoom_region),
+				     BONOBO_OBJECT (zoom_region->properties));
 	zoom_region->timing_output = FALSE;
-	g_timeout_add_full (G_PRIORITY_DEFAULT_IDLE,
-			    200,
-			    zoom_region_update_pointer_timeout,
-			    zoom_region,
-			    NULL);
+#ifdef ZOOM_REGION_DEBUG
+	zoom_region->alive = TRUE;
+#endif
+	zoom_region->priv->update_pointer_id =
+	    g_timeout_add_full (G_PRIORITY_DEFAULT_IDLE,
+				200,
+				zoom_region_update_pointer_timeout,
+				zoom_region,
+				NULL);
 }
 
 ZoomRegion *
@@ -2950,26 +3152,43 @@ zoom_region_new (void)
 	return g_object_new (zoom_region_get_type(), NULL);
 }
 
+/* this one really shuts down the object - called once only */
 static void
 zoom_region_finalize (GObject *region)
 {
 	ZoomRegion *zoom_region = (ZoomRegion *) region;
 
-	g_list_free (zoom_region->priv->q);
-	gtk_container_remove (GTK_CONTAINER (zoom_region->priv->border),
-			      GTK_WIDGET (zoom_region->priv->w));
-	gtk_container_remove (GTK_CONTAINER (((Magnifier *) 
-					      zoom_region->priv->parent)->priv->canvas),
-			      GTK_WIDGET (zoom_region->priv->border));
-	g_object_unref (zoom_region->priv->source_pixbuf_cache);
-	g_object_unref (zoom_region->priv->scaled_pixbuf);
-	g_object_unref (zoom_region->priv->pixmap);
+	DBG(g_message ("finalizing region %p", zoom_region));
+
+	if (zoom_region->priv && zoom_region->priv->q) 
+	{
+		g_list_free (zoom_region->priv->q);
+		zoom_region->priv->q = NULL;
+	}
+	if (GTK_IS_WIDGET (zoom_region->priv->w))
+		gtk_container_remove (GTK_CONTAINER (zoom_region->priv->border),
+				      GTK_WIDGET (zoom_region->priv->w));
+	if (GTK_IS_WIDGET (zoom_region->priv->border))
+		gtk_container_remove (GTK_CONTAINER (((Magnifier *) 
+						      zoom_region->priv->parent)->priv->canvas),
+				      GTK_WIDGET (zoom_region->priv->border));
+	if (zoom_region->priv->source_pixbuf_cache) 
+	    g_object_unref (zoom_region->priv->source_pixbuf_cache);
+	if (zoom_region->priv->scaled_pixbuf) 
+	    g_object_unref (zoom_region->priv->scaled_pixbuf);
+	if (zoom_region->priv->pixmap) 
+	    g_object_unref (zoom_region->priv->pixmap);
 	zoom_region->priv->pixmap = NULL;
-	g_object_unref (zoom_region->priv->cursor_backing_pixels);
-	g_object_unref (zoom_region->priv->border_gc);
+	zoom_region->priv->parent = NULL;
+	if (zoom_region->priv->cursor_backing_pixels) 
+	    g_object_unref (zoom_region->priv->cursor_backing_pixels);
+	if (zoom_region->priv->border_gc) 
+	    g_object_unref (zoom_region->priv->border_gc);
 	g_free (zoom_region->priv);
 	zoom_region->priv = NULL;
-
+#ifdef ZOOM_REGION_DEBUG
+	zoom_region->alive = FALSE;
+#endif
 	BONOBO_CALL_PARENT (G_OBJECT_CLASS, finalize, (region));
 }
 
